"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_cv.src.metrics.object_detection.box_coco_metrics import BoxCOCOMetrics
