"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_cv.datasets.pascal_voc import segmentation
from keras_cv.src.datasets.pascal_voc.load import load
